<?php

namespace Drupal\micro_simple_sitemap\Batch;

use Drupal\Core\Url;
use Drupal\micro_site\Entity\Site;
use Drupal\simple_sitemap\Batch\EntityUrlGenerator as SimpleSitemapEntityUrlGenerator;

/**
 * Class EntityUrlGenerator
 * @package Drupal\micro_simple_sitemap\Batch
 */
class EntityUrlGenerator extends SimpleSitemapEntityUrlGenerator {

  /**
   * {@inheritdoc}
   */
  public function generate($entity_info) {

    foreach ($this->getBatchIterationElements($entity_info) as $entity_id => $entity) {

      $this->setCurrentId($entity_id);

      $entity_settings = $this->generator->getEntityInstanceSettings($entity_info['entity_type_name'], $entity_id);

      if (empty($entity_settings['index'])) {
        continue;
      }

      switch ($entity_info['entity_type_name']) {
        // Loading url object for menu links.
        case 'menu_link_content':
          if (!$entity->isEnabled()) {
            continue 2;
          }
          $url_object = $entity->getUrlObject();
          break;

        // Loading url object for other entities.
        default:
          $url_object = $entity->toUrl();
      }

      // Do not include external paths.
      if (!$url_object->isRouted()) {
        continue;
      }

      $path = $url_object->getInternalPath();

      if ($this->pathSiteProcessed($entity_info['site_id'], $path)) {
        continue;
      }

      $url_object->setOption('absolute', TRUE);

      $path_data = [
        'path' => $path,
        'entity_info' => [
          'entity_type' => $entity_info['entity_type_name'],
          'id' => $entity_id
        ],
        'lastmod' => method_exists($entity, 'getChangedTime')
          ? date_iso8601($entity->getChangedTime()) : NULL,
        'priority' => isset($entity_settings['priority']) ? $entity_settings['priority'] : NULL,
        'changefreq' => !empty($entity_settings['changefreq']) ? $entity_settings['changefreq'] : NULL,
        'site_id' => $entity_info['site_id'],
      ];

      $this->addUrl($path_data, $url_object);
    }
    $this->processSegment();
  }

  /**
   * {@inheritdoc}
   */
  protected function getBatchIterationElements(array $entity_info) {
    $query = $this->entityTypeManager->getStorage($entity_info['entity_type_name'])->getQuery();

    if (!empty($entity_info['keys']['id'])) {
      $query->sort($entity_info['keys']['id'], 'ASC');
    }
    if (!empty($entity_info['keys']['bundle'])) {
      if ($entity_info['entity_type_name'] == 'menu_link_content') {
        if ($entity_info['site_id']) {
          $menu_name = Site::load($entity_info['site_id'])->getSiteMenu();
          $query->condition($entity_info['keys']['bundle'], $menu_name);
        }
        else {
          // Default to main
          // @TODO exclude all menu attached to site entities.
          $query->condition($entity_info['keys']['bundle'], 'main');
        }
      }
      else {
        $query->condition($entity_info['keys']['bundle'], $entity_info['bundle_name']);
      }

    }
    if (!empty($entity_info['keys']['status'])) {
      $query->condition($entity_info['keys']['status'], 1);
    }
    if ($entity_info['entity_type_name'] == 'node') {
      if ($entity_info['site_id']) {
//        $query->condition('site_id', $entity_info['site_id']);
        // @TODO What doing whith nodes on mutliple sites if they can have
        // multiple canonical urls ?
        $orGroupDomain = $query->orConditionGroup()
          ->condition('site_id', $entity_info['site_id'])
          ->condition('field_sites.target_id', $entity_info['site_id'], 'IN')
          ->condition('field_sites_all', 1);
        $query->condition($orGroupDomain);
      }
      else {
        $query->condition('site_id', NULL, 'IS NULL');
      }

    }

    if ($this->needsInitialization()) {
      $count_query = clone $query;
      $this->initializeBatch($count_query->count()->execute());
    }

    if ($this->isBatch()) {
      $query->range($this->context['sandbox']['progress'], $this->batchInfo['batch_process_limit']);
    }

    return $this->entityTypeManager
      ->getStorage($entity_info['entity_type_name'])
      ->loadMultiple($query->execute());
  }

  /**
   * {@inheritdoc}
   */
  protected function addUrl(array $path_data, Url $url_object = NULL) {
    if ($url_object !== NULL) {
      $site = Site::load($path_data['site_id']);
      $micro_site_config = \Drupal::config('micro_site.settings');
      $master_path = $micro_site_config->get('base_scheme') . '://' . $micro_site_config->get('base_url');
      $site_path = ($site) ? $site->getSitePath() : $master_path;

      $url_object->setOption('base_url', $site_path);
      $this->addUrlVariants($path_data, $url_object);
    }
    else {
      $this->context['results']['generate'][] = $path_data;
    }
  }

  /**
   * @param integer $site_id
   * @param string $path
   * @return bool
   */
  protected function pathSiteProcessed($site_id, $path) {
    $path_pool = isset($this->context['results']['processed_site_paths'][$site_id]) ? $this->context['results']['processed_site_paths'][$site_id] : [];
    if (in_array($path, $path_pool)) {
      return TRUE;
    }
    $this->context['results']['processed_site_paths'][$site_id][] = $path;
    return FALSE;
  }

}
