<?php

namespace Drupal\micro_simple_sitemap\Controller;

use Drupal\micro_simple_sitemap\MicroSimpleSitemap;
use Drupal\Core\Cache\CacheableResponse;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Database\Connection;
use Drupal\domain\DomainNegotiator;
use Drupal\micro_site\SiteNegotiatorInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Class SimplesitemapController.
 *
 * @package Drupal\simple_sitemap\Controller
 */
class MicroSimpleSitemapController extends ControllerBase {

  /**
   * The sitemap generator.
   *
   * @var \Drupal\micro_simple_sitemap\MicroSimpleSitemap
   */
  protected $generator;

  /**
   * The micro site negotiator.
   *
   * @var \Drupal\micro_site\SiteNegotiatorInterface
   */
  protected $negotiator;

  /**
   * Database Connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $db;

  /**
   * MicroSimpleSitemapController constructor.
   *
   * @param \Drupal\micro_simple_sitemap\MicroSimpleSitemap $generator
   *   The sitemap generator.
   * @param \Drupal\micro_site\SiteNegotiatorInterface $site_negotiator
   *   The Domain Negotiator.
   * @param \Drupal\Core\Database\Connection $db
   *   The Database Connection.
   */
  public function __construct(MicroSimpleSitemap $generator, SiteNegotiatorInterface $site_negotiator, Connection $db) {
    $this->generator = $generator;
    $this->negotiator = $site_negotiator;
    $this->db = $db;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('micro_simple_sitemap.generator'),
      $container->get('micro_site.negotiator'),
      $container->get('database')
    );
  }

  /**
   * Returns the whole sitemap, requested sitemap chunk or sitemap index file.
   *
   * @param int $chunk_id
   *   Optional ID of the sitemap chunk. If none provided, the first chunk or
   *   the sitemap index is fetched.
   *
   * @throws NotFoundHttpException
   *
   * @return object
   *   Returns an XML response.
   */
  public function getSitemap($chunk_id = NULL) {
    // Get active site.
    $site = $this->negotiator->getActiveSite();
    $site_id = ($site) ? $site->id() : 0;
    // Get chunk of the sitemap by domain and display it.
    $chunk_id = $this->db->query('SELECT id FROM {simple_sitemap} WHERE site_id = :site_id', [':site_id' => $site_id])->fetchField();
    if (!$chunk_id) {
      $chunk_id = NULL;
    }
    $output = $this->generator->getSitemap($chunk_id);
    if (!$output) {
      throw new NotFoundHttpException();
    }

    $response = new CacheableResponse('', 200);
    $response->setContent($output);
    $cache_metadata = $response->getCacheableMetadata();
    $cache_metadata->addCacheTags(['simple_sitemap:' . $site_id]);
    $cache_metadata->addCacheContexts(['url.site']);
    $response->addCacheableDependency($cache_metadata);

    $response->headers->set('Content-type', 'application/xml');

    return $response;
  }
}
